import time
from typing import Dict, List, Tuple

import requests

from mwconstants.constants.c_languages import WIKIPEDIA_LANGUAGES
from mwconstants.constants.c_namespaces import DEFAULT_NAMESPACES, NAMESPACE_ALIASES
from mwconstants.languages import get_wiki_sites


def get_all_namespace_prefix_map(
    lang: str, project: str = "wikipedia"
) -> Dict[str, dict]:
    """Get mapping of namespaces to canonical / possible aliases for a given wiki.

    For return dictionary:
    * Keys are the local names -- i.e. the official names used in parsed article HTML
    * Values include the canonical name -- e.g., User for fr:Utilisateur
    * Values include the list of possible aliases -- e.g., Utilisatrice
    * Together, the key + canonical + aliases make all the valid prefixes for a namespace

    More details: https://www.mediawiki.org/wiki/Help:Namespaces#Localisation

    NOTE: this data can alternatively be extracted from another dump:
    <lang>-<date>-siteinfo-namespaces.json.gz
    """
    session = requests.Session()
    base_url = f"https://{lang}.{project}.org/w/api.php"
    params = {
        "action": "query",
        "meta": "siteinfo",
        "siprop": "namespaces|namespacealiases",
        "format": "json",
        "formatversion": "2",
    }
    result = session.get(url=base_url, params=params).json()

    prefixes = {}
    id_to_name = {}  # e.g., 1 -> Talk; needed for namespace aliases
    if "namespaces" in result.get("query", {}):
        for ns in result["query"]["namespaces"].values():
            try:
                canonical = ns["canonical"]
                name = ns["name"]
                nid = ns["id"]
                id_to_name[nid] = name
                prefixes[name] = {"canonical": canonical}
            except KeyError:  # main namespace has no canonical prefix and we want to skip it
                continue
    if "namespacealiases" in result.get("query", {}):
        for alias in result["query"]["namespacealiases"]:
            try:
                nid = alias["id"]
                prefix = alias["alias"]
                name = id_to_name[nid]
                prefixes[name]["aliases"] = prefixes[name].get("aliases", []) + [prefix]
            except KeyError:
                print(f"Invalid alias: {alias}")

    return prefixes


def gather_all_languages(
    deduplicate: bool = True,
) -> Tuple[Dict[str, List[str]], Dict[str, dict]]:
    """Helper function for gathering namespace data of a given type."""
    wiki_languages = get_wiki_sites("wiki")
    print(f"{len(wiki_languages)} languages: {wiki_languages}")
    namespaces = {}
    for lang in wiki_languages:
        namespaces[lang] = get_all_namespace_prefix_map(lang, "wikipedia")
        time.sleep(0.25)

    def_ns = {}
    # make sure to copy values not pointers so it's not later removed with English
    for ns_name in namespaces["en"]:
        canonical = namespaces["en"][ns_name]["canonical"]
        if ns_name == canonical:
            aliases = list(namespaces["en"][ns_name].get("aliases", []))
        else:
            aliases = list(
                set(namespaces["en"][ns_name].get("aliases", []) + [ns_name])
            )
        def_ns[canonical] = aliases

    if deduplicate:
        langs_to_remove = []
        for lang in namespaces:
            ns_to_remove = []
            for ns_name in namespaces[lang]:
                ns_prefixes = namespaces[lang][ns_name]
                canonical = ns_prefixes["canonical"]
                if canonical not in def_ns:
                    # Special namespaces that don't appear in English: remove duplicate canonical if relevant
                    if ns_name == canonical:
                        ns_prefixes.pop("canonical")
                    continue
                # remove official aliases that match any default prefixes (canonical/aliases)
                def_prefixes = [canonical] + def_ns[canonical]
                for i in range(len(ns_prefixes.get("aliases", [])) - 1, -1, -1):
                    if ns_prefixes["aliases"][i] in def_prefixes:
                        ns_prefixes["aliases"].pop(i)
                if "aliases" in ns_prefixes and not ns_prefixes["aliases"]:
                    ns_prefixes.pop("aliases")
                # only keep canonical if different from name
                if ns_name == canonical:
                    ns_prefixes.pop("canonical")
                if not ns_prefixes:
                    ns_to_remove.append(ns_name)

            for ns_name in ns_to_remove:
                namespaces[lang].pop(ns_name)
            if not namespaces[lang]:
                langs_to_remove.append(lang)

        for lang in langs_to_remove:
            namespaces.pop(lang)

    return def_ns, namespaces


def standardize_prefix(prefix: str) -> str:
    """Standardize link format"""
    return prefix.split(":", maxsplit=1)[0].replace(" ", "_").capitalize()


def prefixes_to_canonical(
    lang: str = "en", interlanguage: bool = True
) -> Dict[str, str]:
    """Build map of prefixes to their canonical namespace prefix.

    NOTE: the use of standardize_prefix is to ensure look-ups in this dictionary are successful
    but it also makes the dictionary look funny because of the enforced capitalize/underscore style
    """
    if interlanguage:
        prefixes = {
            standardize_prefix(wiki_lang): "Interlanguage"
            for wiki_lang in WIKIPEDIA_LANGUAGES
            if wiki_lang != lang
        }
    else:
        prefixes = {}
    for canonical_ns in DEFAULT_NAMESPACES:
        prefixes[standardize_prefix(canonical_ns)] = canonical_ns
        for alias in DEFAULT_NAMESPACES[canonical_ns]:
            prefixes[standardize_prefix(alias)] = canonical_ns

    for ns in NAMESPACE_ALIASES.get(lang, {}):
        canonical = NAMESPACE_ALIASES[lang][ns].get("canonical", ns)
        prefixes[standardize_prefix(ns)] = canonical
        prefixes[standardize_prefix(canonical)] = canonical
        for alias in NAMESPACE_ALIASES[lang][ns].get("aliases", []):
            prefixes[standardize_prefix(alias)] = canonical

    return prefixes


def get_language_list_prefixes(namespace: str, lang: str) -> List[str]:
    """Generate a list of acceptable prefixes for a particular language+namespace.

    For example: get_language_list_prefixes('File', lang='fr') -> ['Fichier', 'File', 'Image']
    Most useful when working with wikitext where these prefixes would all be valid.
    If working with parsed HTML, only the local language prefix will be used. If checking
    for containment, make sure to standardize both the prefixes and input e.g., via standardize_prefix
    so that e.g., FILE correctly matches against File.
    """

    # try to canonicalize the namespace
    if namespace not in DEFAULT_NAMESPACES:
        for lang_ns in NAMESPACE_ALIASES.get(lang, {}):
            # matches local name, canonical, or an alias
            if (
                namespace == lang_ns
                or namespace == NAMESPACE_ALIASES[lang][lang_ns].get("canonical")
                or namespace in NAMESPACE_ALIASES[lang][lang_ns].get("aliases", [])
            ):
                namespace = NAMESPACE_ALIASES[lang][lang_ns].get("canonical", lang_ns)

    # extract default namespace prefixes
    lang_ns_prefixes = []
    if namespace in DEFAULT_NAMESPACES:
        lang_ns_prefixes.append(namespace)
        lang_ns_prefixes.extend(DEFAULT_NAMESPACES[namespace])

    # add language-specific prefixes
    for lang_ns in NAMESPACE_ALIASES.get(lang, {}):
        if (
            lang_ns == namespace
            or NAMESPACE_ALIASES[lang][lang_ns].get("canonical", None) == namespace
        ):
            lang_ns_prefixes.append(lang_ns)
            canonical_ns = NAMESPACE_ALIASES[lang][lang_ns].get("canonical", None)
            if canonical_ns:
                lang_ns_prefixes.append(canonical_ns)
            lang_ns_prefixes.extend(NAMESPACE_ALIASES[lang][lang_ns].get("aliases", []))

    if not lang_ns_prefixes:
        print(
            f"No results. Maybe because namespace ({namespace}) non-standard: {list(DEFAULT_NAMESPACES.keys())}"
        )
        return []
    else:
        return list(set(lang_ns_prefixes))


if __name__ == "__main__":
    (
        default_namespaces,
        namespace_aliases,
    ) = gather_all_languages()  # Wikipedia namespaces
    print(default_namespaces)
    print(namespace_aliases)



import copy
import time
from typing import Dict, Tuple

import requests

from mwconstants.constants.c_media import DEF_OPTION_TAGS, IMG_OPTION_ALIASES
from mwconstants.languages import get_wiki_sites


def parse_image_options(wikitext: str, lang: str = "en") -> Tuple[str, str, list]:
    """Split media file wikitext into file title, formatting options, and caption.

    A media link always begins with the title and any formatting/captions follow as parameters separated by |.
    Wikitext media can show up in three different formats, each of which is supported here:
    * Brackets: [[File:filename.ext|formatting options|caption]]
    * Template: File:filename.ext (caption/formatting are separate parameters)
    * Gallery: filename.ext|formatting options|caption

    There are a set of allowed keywords/parameters for media formatting. Each parameter in a media file is
    checked against these formatting options and if it does not match, it is assumed to be the caption.

    For more info, see: https://www.mediawiki.org/wiki/Help:Images#Syntax
    """
    title = None
    options = []
    caption = None
    if wikitext:
        # remove optional brackets and split into parameters
        wikitext_parts = wikitext.strip("[]").split("|")
        # title is the first parameter and the only required one
        title = wikitext_parts[0]
        if len(wikitext_parts) > 1:
            lang_tags = {}
            for k in DEF_OPTION_TAGS:
                lang_tags[k] = DEF_OPTION_TAGS[k] + IMG_OPTION_ALIASES.get(
                    lang, {}
                ).get(k, [])
            for o in wikitext_parts[1:]:
                o = o.strip()
                if o in lang_tags["keywords"]:
                    options.append(o)
                elif o.split("=")[0] in lang_tags["params"]:
                    options.append(o)
                else:
                    found = False
                    for t in lang_tags["startswith"]:
                        if o.startswith(t):
                            options.append(o)
                            found = True
                            break
                    if not found:
                        for t in lang_tags["endswith"]:
                            if o.endswith(t):
                                options.append(o)
                                found = True
                                break
                    if not found:
                        caption = o

    return title, caption, options


def get_wiki_valid_img_options(lang: str) -> Dict[str, list]:
    """Get official list of acceptable image formatting tags -- e.g., frameless, top, etc."""
    session = requests.Session()
    base_url = f"https://{lang}.wikipedia.org/w/api.php"
    params = {
        "action": "query",
        "meta": "siteinfo",
        "siprop": "magicwords",
        "format": "json",
        "formatversion": "2",
    }
    result = session.get(url=base_url, params=params).json()

    img_keywords = []
    img_param_names = []
    img_startswith = []
    img_endswith = []
    # TODO: carry magic["name"] over with the parameters and
    #  map them to a more fine-grained taxonomy than just options
    if "magicwords" in result.get("query", {}):
        for magic in result["query"]["magicwords"]:
            if magic["name"].startswith("img_"):
                if not magic["case-sensitive"]:
                    # we don't handle case insensitive so...
                    print(f"warning case-insensitive for {lang}: {magic}")
                for tag in magic["aliases"]:
                    if "=" in tag:
                        img_param_names.append(tag.split("=")[0])
                    elif tag.endswith("$1"):
                        img_startswith.append(tag[:-2])
                    elif tag.startswith("$1"):
                        img_endswith.append(tag[2:])
                    elif "$" in tag:
                        print("Didn't expect this:", tag)
                    else:
                        img_keywords.append(tag)
    return {
        "keywords": list(set(img_keywords)),
        "params": list(set(img_param_names)),
        "startswith": list(set(img_startswith)),
        "endswith": list(set(img_endswith)),
    }


def get_img_options(project: str = "wiki") -> Tuple[dict, Dict[str, dict]]:
    """
    Utility for generating language-specific lists of valid media formatting options on a Wikimedia project.

    For example, to know that [[File:image.png|thumb]] has `thumb` formatting as opposed to a caption
    that reads `thumb` we must know that `thumb` is a valid image formatting tag and appears as a
    stand-alone keyword. Any parameters that do not match a valid image formatting tag pattern are
    assumed to be the image caption.

    See: https://www.mediawiki.org/wiki/Help:Images#Syntax
    """

    wiki_languages = get_wiki_sites(project)
    print(f"{len(wiki_languages)} languages: {wiki_languages}")
    img_options = {}
    for lang in wiki_languages:
        img_options[lang] = get_wiki_valid_img_options(lang)
        time.sleep(0.25)

    # separate out english tags, which are valid for any language
    # this greatly reduces the size of the final list
    default_option_tags = copy.deepcopy(img_options["en"])

    to_remove = []
    for lang in img_options:
        for param_type in default_option_tags:
            for tag in default_option_tags[param_type]:
                img_options[lang][param_type].remove(tag)
            if not img_options[lang][param_type]:
                img_options[lang].pop(param_type)
        if not img_options[lang]:
            to_remove.append(lang)

    for lang in to_remove:
        img_options.pop(lang)

    return default_option_tags, img_options


if __name__ == "__main__":
    print(get_img_options("wiki"))  # Wikipedia image formatting options



import codecs
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = "0.1.0"
DESCRIPTION = "Various data and utilities for processing wikitext."

# Dev dependencies
EXTRAS_REQUIRE = {
    "tests": ["pytest>=6.2.5"],
}

EXTRAS_REQUIRE["dev"] = EXTRAS_REQUIRE["tests"]

# Setting up
setup(
    name="mwconstants",
    version=VERSION,
    author="geohci (Isaac Johnson)",
    author_email="<isaac@wikimedia.org>",
    url="https://gitlab.wikimedia.org/repos/research/mwconstants",
    license="MIT License",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=["requests"],
    keywords=["python", "wikitext", "wiki"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    extras_require=EXTRAS_REQUIRE,
    include_package_data=True,
    zip_safe=False,
)
